#!/usr/bin/env python3
"""

"""
# ruff: noqa: ANN001, ANN202, B011, F841, ARG002

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import logging as logmod
import pathlib as pl
import warnings
from collections import ChainMap
# ##-- end stdlib imports

# ##-- 3rd party imports
import pytest

# ##-- end 3rd party imports

# ##-- 1st party imports
from jgdv import identity_fn
# ##-- end 1st party imports

from .._interface import ExpInst_d, SourceChain_d, ExpInstChain_d, InstructionFactory_p, IndirectKey_p, EXPANSION_CONVERT_MAPPING
from ... import DKey, IndirectDKey, NonDKey
from ..expander_stack import DKeyExpanderStack, InstructionFactory

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from collections.abc import Mapping

if TYPE_CHECKING:
   from jgdv import Maybe
   from typing import Final
   from typing import ClassVar, Any, LiteralString
   from typing import Never, Self, Literal
   from typing import TypeGuard
   from collections.abc import Iterable, Iterator, Callable, Generator
   from collections.abc import Sequence, MutableMapping, Hashable

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
logmod.getLogger("jgdv.structs.dkey._util.expander_stack").setLevel(logmod.INFO)
##-- end logging

# Vars:
Expanders : Final[list[type]] = [DKeyExpanderStack]
# Body:

class TestExpInst_d:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        obj = ExpInst_d(value="blah", fallback="bloo")
        assert(obj.value == "blah")
        assert(obj.rec is None)
        assert(obj.fallback == "bloo")

    def test_no_val_errors(self):
        with pytest.raises(KeyError):
            ExpInst_d(fallback="bloo")

    def test_match(self):
        match ExpInst_d(value="blah", fallback="bloo"):
            case ExpInst_d(value="blah"):
                assert(True)
            case x:
                assert(False), x

    def test_match_fail(self):
        match ExpInst_d(value="bloo", fallback="bloo"):
            case ExpInst_d(rec=True):
                assert(False)
            case _:
                assert(True)

class TestExpInstChain_d:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

class TestSourceChain_d:

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self):
        match SourceChain_d():
            case SourceChain_d():
                assert(True)
            case x:
                assert(False), x

    def test_with_base(self):
        match SourceChain_d({"a":2, "b":3}):
            case SourceChain_d():
                assert(True)
            case x:
                assert(False), x

    def test_with_specstruct_base(self):

        class SimpleSpecStruct:

            @property
            def params(self) -> dict:
                return {"blah":"aweg"}

            @property
            def args(self) -> list:
                return []

            @property
            def kwargs(self) -> dict:
                return {}

        match SourceChain_d(SimpleSpecStruct()):
            case SourceChain_d() as obj:
                assert(obj.get("blah") == "aweg")
            case x:
                assert(False), x

    def test_with_multi_base(self):
        match SourceChain_d({"a":2, "b":3}, {"blah":"bloo"}):
            case SourceChain_d():
                assert(True)
            case x:
                assert(False), x

    def test_extend(self):
        obj = SourceChain_d({"a":2, "b":3}, {"blah":"bloo"})
        match obj.extend({"blee":"aweg"}):
            case SourceChain_d() as extended:
                assert(obj is not extended)
                assert(extended.get("blee") == "aweg")
            case x:
                assert(False), x

    def test_simple_get(self):
        obj = SourceChain_d({"a":2, "b":3})
        match obj.get("b"):
            case 3:
                assert(True)
            case x:
                assert(False), x

    def test_simple_get_with_fallback(self):
        obj = SourceChain_d({"a":2, "b":3})
        match obj.get("d", 10):
            case 10:
                assert(True)
            case x:
                assert(False), x

    def test_multi_base_get(self):
        obj = SourceChain_d({"a":2, "b":3}, {"blah":"bloo"})
        match obj.get("blah"):
            case "bloo":
                assert(True)
            case x:
                assert(False), x

    def test_lookup(self):
        obj   = SourceChain_d({"a":2, "b":3}, {"blah":"bloo"})
        inst  = ExpInstChain_d(ExpInst_d(value="blah"),
                               root=DKey("blah"))
        match obj.lookup(inst):
            case ("bloo", x) if x == inst.chain[0]:
                assert(True)
            case x:
                assert(False), x

class TestInstructionFactory:

    @pytest.fixture(scope="function")
    def fac(self):
        return InstructionFactory(ctor=DKey)

    def test_sanity(self):
        assert(True is not False) # noqa: PLR0133

    def test_match_protocol(self, fac):
        assert(isinstance(fac, InstructionFactory_p))

    def test_null_inst(self, fac):
        match fac.null_inst():
            case ExpInst_d(value=None):
                assert(True)
            case x:
                assert(False), x

    def test_simple_inst(self, fac):
        match fac.build_inst(DKey("{test}"), None, {}):
            case ExpInst_d(value="test"):
                assert(True)
            case x:
                assert(False), x

    def test_list_val_inst(self, fac):
        match fac.build_inst([1,2,3,4], None, {}):
            case ExpInst_d(value=[1,2,3,4], literal=True):
                assert(True)
            case x:
                assert(False), x

    def test_lift_implicit(self, fac):
        """ Because the root is indirect, the subsequent inst is lifted """
        root = fac.build_inst(DKey("{test_}"), None, {})
        match fac.build_inst("blah", root, {}):
            case ExpInst_d(value=DKey() as x, literal=False) if x == "blah":
                assert(True)
            case x:
                assert(False), x

    def test_prefer_key_hook(self, fac):

        class CustomHookKey(DKey):
            __slots__ = ()

            def exp_to_inst_h(self, root, factor, **kwargs):
                return ExpInst_d(value="test built", literal=True)

        assert(hasattr(CustomHookKey, "exp_to_inst_h"))
        key = DKey("{blah}", force=CustomHookKey)
        assert(isinstance(key, CustomHookKey))
        assert(hasattr(key, "exp_to_inst_h"))

        match fac.build_inst(key, None, {}):
            case ExpInst_d(value="test built", literal=True):
                assert(True)
            case x:
                assert(False), x

    def test_auto_lift_to_key(self, fac):
        match fac.build_inst("{test}", None, {}):
            case ExpInst_d(value=DKey() as x) if x == "test":
                assert(True)
            case x:
                assert(False), x

    def test_decrement_recursion(self, fac):
        initial = fac.build_inst(DKey("{test:e3}"), None, {}, decrement=False)
        assert(initial.rec == 3)
        match fac.build_inst(DKey("{test}"), initial, {}):
            case ExpInst_d(value="test", rec=2):
                assert(True)
            case x:
                assert(False), x

    def test_decrement_recursion_zero(self, fac):
        initial = fac.build_inst(DKey("{test:e0}"), None, {})
        assert(initial.rec == 0)
        match fac.build_inst(DKey("{test}"), initial, {}):
            case ExpInst_d(value="test", rec=0):
                assert(True)
            case x:
                assert(False), x

    def test_decrement_recursion_null(self, fac):
        initial = fac.build_inst(DKey("{test}"), None, {})
        assert(initial.rec is None)
        match fac.build_inst(DKey("{blah}"), initial, {}):
            case ExpInst_d(value="blah", rec=None):
                assert(True)
            case x:
                assert(False), x

    def test_decrement_recursion_guard(self, fac):
        initial = fac.build_inst(DKey("{test}"), None, {})
        assert(initial.rec is None)
        match fac.build_inst(DKey("{test}"), initial, {}):
            case ExpInst_d(value="test", rec=10):
                assert(True)
            case x:
                assert(False), x

    def test_basic_chain(self, fac):
        base = fac.build_inst("{blah}", None, {})
        match fac.build_chains(base, {}):
            case [ExpInstChain_d() as chain]:
                assert(len(chain) == 3)
                assert(chain[-1].value is None)
                assert(True)
            case x:
                assert(False), x

    def test_chains_dont_decrement_rec(self, fac):
        base = fac.build_inst("{blah:e5}", None, {}, decrement=False)
        assert(base.rec == 5)
        match fac.build_chains(base, {}):
            case [ExpInstChain_d() as chain]:
                assert(len(chain) == 3)
                assert(all(x.rec == 5 for x in chain[:-1]))
                assert(chain[-1].value is None)
                assert(True)
            case x:
                assert(False), x

    def test_custom_chain_hook(self, fac):

        class CustomChainHook(DKey):
            __slots__ = ()

            def exp_generate_chains_h(self, val, fac, opts):
                return [ExpInstChain_d(root=val.value)]

        key = DKey("{blah}", force=CustomChainHook)
        assert(isinstance(key, CustomChainHook))
        base = fac.build_inst(key, None, {})
        match fac.build_chains(base, {}):
            case [ExpInstChain_d() as chain]:
                assert(len(chain) == 0)
                assert(chain.root == key)
            case x:
                assert(False), x

    def test_chain_order(self, fac):
        """ Default order is [{key:d}, {key:i}, None] """

        key   = DKey("{blah}")
        base  = fac.build_inst(key, None, {})
        match fac.build_chains(base, {}):
            case [ExpInstChain_d() as chain]:
                assert(len(chain) == 3)
                assert(chain[0].value == key)
                assert(chain[1].value is not key)
                assert(chain[1].value == "blah")
                assert(isinstance(chain[1].value, IndirectKey_p))
                assert(chain.root is key)
            case x:
                assert(False), x

    def test_indirect_chain(self, fac):
        """ Indirect Keys swap the order to [{key:i}, {key:d}, None] """

        key   = DKey("{blah_}")
        base  = fac.build_inst(key, None, {})
        match fac.build_chains(base, {}):
            case [ExpInstChain_d() as chain]:
                assert(len(chain) == 3)
                assert(chain[0].value is key)
                assert(str(chain[0].value) == "blah_")
                assert(isinstance(chain[0].value, IndirectKey_p))
                assert(str(chain[1].value) == "blah")
                assert(chain.root is key)
            case x:
                assert(False), x

@pytest.mark.parametrize("ctor", Expanders)
class TestExpander:

    @pytest.fixture(scope="function")
    def source(self, mocker):
        return SourceChain_d()

    @pytest.fixture(scope="function")
    def exp(self, ctor):
        return ctor(ctor=DKey)

    def test_sanity(self, exp):
        assert(True is not False) # noqa: PLR0133

    def test_factory_access(self, ctor):
        match ctor():
            case ctor() as x:
                assert(isinstance(x._factory, InstructionFactory))
            case x:
                assert(False), x

    def test_set_ctor(self, ctor):
        obj = ctor()
        obj._factory._ctor = None
        obj.set_ctor(DKey)
        assert(obj._factory._ctor is DKey)

    def test_extra_sources(self, exp):

        class SimpleDKey(DKey):
            __slots__ = ()

            @override
            def exp_extra_sources_h(self, sources:SourceChain_d) -> SourceChain_d:
                return sources.extend([1,2,3,4])

        key = ExpInstChain_d(root=DKey("blah", force=SimpleDKey))
        assert(isinstance(key.root, SimpleDKey))
        match exp.extra_sources(key, SourceChain_d()):
            case SourceChain_d() as x:
                assert(x.sources[0] == [1,2,3,4])
            case x:
                assert(False), x

    def test_do_lookup(self, exp, source):
        key = ExpInstChain_d(ExpInst_d(value="simple"),
                             root=DKey("{simple}"),
                             )
        ext = source.extend({"simple":"blah"})
        match exp.do_lookup(key, ext, {}):
            case ExpInst_d(value="blah"):
                assert(True)
            case x:
                assert(False), x

    def test_flatten(self, exp, source):
        obj      = exp
        obj.set_ctor(DKey)
        root = DKey("{simple}")
        vals = [
            ExpInst_d(value="simple"),
        ]
        match obj.flatten(vals, root, {}):
            case ExpInst_d() as x if x is vals[0]:
                assert(True)
            case x:
                assert(False), x

    def test_coerce_result_no_op(self, exp):
        obj      = exp
        obj.set_ctor(DKey)
        root     = DKey("{simple}")
        inst     = ExpInst_d(value="simple", rec=1)
        sources  = [{"simple":"blah"}]
        match obj.coerce_result(inst, root, {}):
            case ExpInst_d(value=str(), literal=True) as x:
                assert(True)
            case x:
                assert(False), x

    def test_coerce_result_simple(self, exp):
        obj      = exp
        obj.set_ctor(DKey)
        root = DKey("{simple}", ctor=pl.Path)
        inst     = ExpInst_d(value="simple", rec=1)
        sources  = [{"simple":"blah"}]
        match obj.coerce_result(inst, root, {}):
            case ExpInst_d(value=pl.Path(), literal=True) as x:
                assert(True)
            case x:
                assert(False), x

    @pytest.mark.skip
    def test_finalise(self, exp):
        pass

    @pytest.mark.skip
    def test_check_result(self, exp):
        pass

@pytest.mark.parametrize("ctor", Expanders)
class TestExpansion:

    @pytest.fixture(scope="function")
    def exp(self, ctor):
        return ctor(ctor=DKey)

    def test_sanity(self, exp):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self, exp):
        """ {test} -> blah """
        obj = DKey("test", implicit=True)
        state = {"test": "blah"}
        match exp.expand(obj, state):
            case ExpInst_d(value="blah"):
                assert(True)
            case x:
                assert(False), x

    def test_basic_fail(self, exp):
        """ {aweg} -> None """
        obj    = DKey("aweg", implicit=True)
        state  = {"test": "blah"}
        match exp.expand(obj, state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_nonkey_expansion(self, exp):
        """ aweg -> aweg """
        obj = DKey("aweg")
        state = {"test": "blah"}
        match exp.expand(obj, state):
            case ExpInst_d(value="aweg"):
                assert(True)
            case x:
                assert(False), x

    def test_simple_recursive(self, exp):
        """
        {test} -> {blah} -> bloo
        """
        obj = DKey("test", implicit=True)
        state = {"test": "{blah}", "blah": "bloo"}
        match exp.expand(obj, state):
            case ExpInst_d(value="bloo"):
                assert(True)
            case x:
                assert(False), x

    def test_double_recursive(self, exp):
        """
        {test} -> {blah}
        {blah} -> {aweg}/{bloo}
        {aweg}/{bloo} -> qqqq/{aweg}
        qqqq/{aweg} -> qqqq/qqqq
        """
        obj   = DKey("test", implicit=True)
        state = {"test": "{blah}", "blah": "{aweg}/{bloo}", "aweg":"qqqq", "bloo":"{aweg}"}
        match exp.expand(obj, state):
            case ExpInst_d(value="qqqq/qqqq"):
                assert(True)
            case x:
                assert(False), x

    def test_recursive_fail(self, exp):
        """
        {test} -> {blah}
        {blah} -> {aweg}/{bloo}
        {aweg} -> None
        {test} -> None
        """
        obj   = DKey("test", implicit=True)
        state = {"test": "{blah}", "blah": "{aweg}/{bloo}", "bloo":"ajqwoj"}
        match exp.expand(obj, state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_coerce_type(self, exp):
        """ test -> str(blah) -> pl.Path(blah) """
        obj   = DKey("test", implicit=True, ctor=pl.Path)
        state = {"test": "blah"}
        match exp.expand(obj, state):
            case ExpInst_d(value=pl.Path()):
                assert(True)
            case x:
                assert(False), x

    def test_check_type(self, exp):
        """ {test} -> pl.Path(blah) """
        obj   = DKey("test", implicit=True, ctor=pl.Path)
        assert(f"{obj!s}" == "test")
        state = {"test": pl.Path("blah")}
        match exp.expand(obj, state):
            case ExpInst_d(value=pl.Path()):
                assert(True)
            case x:
                assert(False), x

    def test_recursion_limit(self, exp):
        obj    = DKey("test", implicit=True)
        state  = {"test": "{blah}", "blah":"{aweg}", "aweg": "bloo"}
        match exp.expand(obj, state, limit=1):
            case ExpInst_d(value=DKey() as x) if x == "blah":
                assert(True)
            case x:
                assert(False), x

    def test_expansion_cascade(self, exp):
        """
        {test} -1-> {blah},
        {test} -2-> {aweg}
        {test} -3-> qqqq
        """
        obj   = DKey("test", implicit=True)
        state = {"test": "{blah}", "blah": "{aweg}", "aweg": "qqqq"}
        assert(exp.expand(obj, state, limit=1).value == "blah")
        assert(exp.expand(obj, state, limit=2).value == "aweg")
        match exp.expand(obj, state, limit=3):
            case ExpInst_d(value=NonDKey() as x):
                assert(x == "qqqq")
            case x:
                assert(False), x

    def test_expansion_limit_format_param(self, exp):
        """
        {test:e1} -> state[test:{blah}, blah:{aweg}, aweg:qqqq] -> {bloo}
        """
        obj   = DKey("test:e1", implicit=True)
        state = {"test": "{blah}", "blah": "{aweg}", "aweg": "qqqq"}
        assert(exp.expand(obj, state).value == "blah")

    def test_expansion_limit_format_param_two(self, exp):
        """
        {test:e2} -> state[test:{blah}, blah:{aweg}, aweg:qqqq] -> {aweg}
        """
        obj   = DKey("test:e2", implicit=True)
        state = {"test": "{blah}", "blah": "{aweg}", "aweg": "qqqq"}
        assert(exp.expand(obj, state).value == "aweg")
        assert(isinstance(exp.expand(obj, state).value, DKey))

    @pytest.mark.skip("TODO")
    def test_additional_sources_recurse(self, exp):
        """ see doot test_dkey.TestDKeyExpansion.test_indirect_wrapped_expansion
        """
        assert(False)

    def test_keep_original_type_on_expansion(self, exp):
        """ {test} -> state[test:[1,2,3]] -> [1,2,3] """
        obj   = DKey("test", implicit=True)
        state = {"test": [1,2,3]}
        match exp.expand(obj, state):
            case ExpInst_d(value=list()):
                assert(True)
            case x:
                assert(False), type(x)

@pytest.mark.parametrize("ctor", Expanders)
class TestIndirection:

    @pytest.fixture(scope="function")
    def exp(self, ctor):
        return ctor(ctor=DKey)

    def test_sanity(self, exp):
        assert(True is not False) # noqa: PLR0133

    def test_hit(self, exp):
        """
        {key} -> state[key:value] -> value
        """
        obj   = DKey("test", implicit=True)
        state = {"test": "blah"}
        match exp.expand(obj, state):
            case ExpInst_d(value="blah"):
                assert(True)
            case x:
                assert(False), x

    def test_hit_ignores_indirect(self, exp):
        """
        {key} -> state[key:value, key_:val2] -> value
        """
        obj   = DKey("test", implicit=True)
        state = {"test": "blah", "test_":"aweg"}
        match exp.expand(obj, state):
            case ExpInst_d(value="blah"):
                assert(True)
            case x:
                assert(False), x

    def test_hard_miss(self, exp):
        """
        {key} -> state[] -> None
        """
        obj   = DKey("test", implicit=True)
        state = {}
        match exp.expand(obj, state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_hard_miss_with_call_fallback(self, exp):
        """
        {key} -> state[] -> 25
        """
        obj   = DKey("test", implicit=True)
        state = {}
        match exp.expand(obj, state, fallback=25):
            case ExpInst_d(value=25):
                assert(True)
            case x:
                assert(False), x

    def test_hard_miss_with_ctor_fallback(self, exp):
        """
        {key} -> state[] -> 25
        """
        obj   = DKey("test", fallback=25, implicit=True)
        state = {}
        match exp.expand(obj, state):
            case ExpInst_d(value=25):
                assert(True)
            case x:
                assert(False), x

    def test_hard_miss_prefer_call_fallback_over_ctor(self, exp):
        """
        {key} -> state[] -> 25
        """
        obj   = DKey("test", fallback=10, implicit=True)
        state = {}
        match exp.expand(obj, state, fallback=25):
            case ExpInst_d(value=25):
                assert(True)
            case x:
                assert(False), x

    def test_hard_miss_indirect(self, exp):
        """
        {key_} -> state[] -> {key_}
        """
        obj = DKey("test_", implicit=True)
        assert(DKey.MarkOf(obj) == Mapping)
        state = {"test_": "blah", "blah": 25}
        match exp.expand(obj, state):
            case ExpInst_d(value=25):
                assert(True)
            case x:
                assert(False), x

    def test_soft_miss(self, exp):
        """
        {key} -> state[key_:blah] -> {blah}
        """
        target  = DKey("blah", implicit=True)
        obj     = DKey("test", implicit=True)
        state   = {"test_": "blah"}
        match exp.expand(obj, state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_soft_hit_direct(self, exp):
        """
        {key_} -> state[key:value] -> value
        """
        obj = DKey("test_", implicit=True)
        assert(DKey.MarkOf(obj) is Mapping)
        state = {"test": "blah"}
        match exp.expand(obj, state):
            case ExpInst_d(value="blah"):
                assert(True)
            case x:
                assert(False), x

    def test_soft_hit_indirect(self, exp):
        """
        {key_} -> state[key_:key2, key2:value] -> {value}
        """
        obj = DKey("test_", implicit=True)
        assert(DKey.MarkOf(obj) is Mapping)
        state = {"test_": "blah", "blah":"bloo"}
        match exp.expand(obj, state):
            case ExpInst_d(value="bloo"):
                assert(True)
            case x:
                assert(False), x

    def test_indirect_prefers_indirect_over_direct(self, exp):
        """
        {key_} -> state[key_:value, key:val2] -> {value}
        """
        obj = DKey("test_", implicit=True)
        assert(DKey.MarkOf(obj) is Mapping)
        state = {"test_": "blah", "test": "aweg", "blah":"bloo"}
        match exp.expand(obj, state):
            case ExpInst_d(value="bloo"):
                assert(True)
            case x:
                assert(False), x

@pytest.mark.parametrize("ctor", Expanders)
class TestMultiExpansion:

    @pytest.fixture(scope="function")
    def exp(self, ctor):
        return ctor(ctor=DKey)

    def test_sanity(self, exp):
        assert(True is not False) # noqa: PLR0133

    def test_basic(self, exp):
        obj = DKey("{test} {test}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah"}
        match exp.expand(obj, state):
            case ExpInst_d(value="blah blah"):
                assert(True)
            case x:
                assert(False), x

    def test_basic_fail(self, exp):
        obj = DKey("{test} {aweg}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah"}
        match exp.expand(obj, state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_hard_miss_subkey(self, exp):
        """ {key}/{key2} -> state[key:value} -> None """
        obj = DKey("{test}/{aweg}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah"}
        match exp.expand(obj, state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_soft_miss_subkey(self, exp):
        """ {key}/{key2} -> state[key:val, key2_:key] -> val/val """
        obj = DKey("{test}/{aweg}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah", "aweg_":"test"}
        match exp.expand(obj, state):
            case ExpInst_d(value="blah/blah"):
                assert(True)
            case x:
                assert(False), x

    def test_indirect_subkey(self, exp):
        """ {key}/{key2_} -> state[key:val, key2_:key] -> val/val """
        obj = DKey("{test}/{aweg_}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah", "aweg_":"test"}
        match exp.expand(obj, state):
            case ExpInst_d(value="blah/blah"):
                assert(True)
            case x:
                assert(False), x

    def test_direct_subkey_when_indirect_missing(self, exp):
        """ {key}/{key2_} -> state[key:val, key2:val2] -> val/val2 """
        obj = DKey("{test}/{aweg_}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah", "aweg":"test"}
        match exp.expand(obj, state):
            case ExpInst_d(value="blah/test"):
                assert(True)
            case x:
                assert(False), x

    def test_indirect_subkey_over_direct(self, exp):
        """ {key}/{key2_} -> state[key:val, key3:val2, key2_:key] -> val/val """
        obj = DKey("{test}/{aweg_}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah", "aweg":"test", "aweg_":"test"}
        match exp.expand(obj, state):
            case ExpInst_d(value="blah/blah"):
                assert(True)
            case x:
                assert(False), x

    def test_indirect_miss_subkey_remains(self, exp):
        """ {key}/{key2_} -> state[key:val] ->  """
        obj = DKey("{test}/{aweg_}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah"}
        match exp.expand(obj, state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_multikey_recursion_missing_subkey(self, exp):
        obj = DKey[list]("{test}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "{blah}", "blah": "blah/{aweg_}"}
        match exp.expand(obj, state):
            case None:
                assert(True)
            case x:
                assert(False), x

    def test_multikey_recursion_depth_10(self, exp):
        obj = DKey[list]("{test}")
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "{test}"}
        match exp.expand(obj, state, limit=10):
            case ExpInst_d(value="{test}"):
                assert(True)
            case x:
                assert(False), x

    def test_multikey_individual_subkey_recursion_limit(self, exp):
        """
        Setting recursion limits with :e[digit]
        {test:e2} -> state[test:{blah}, blah:{aweg}, aweg:qqqq] -> {aweg}
        """
        obj = DKey("{test:e1} : {test:e2} : {test:e3}")
        state = {"test": "{blah}", "blah": "{aweg}", "aweg": "qqqq"}
        assert(exp.expand(obj, state).value == "{blah} : {aweg} : qqqq")

@pytest.mark.parametrize("ctor", Expanders)
class TestCoercion:

    @pytest.fixture(scope="function")
    def exp(self, ctor):
        return ctor(ctor=DKey)

    def test_sanity(self, exp):
        assert(True is not False) # noqa: PLR0133

    def test_coerce_param_path(self, exp):
        obj = DKey("{test!p}")
        state = {"test": "blah"}
        assert(obj.data.convert == "p")
        match exp.expand(obj, state):
            case ExpInst_d(value=pl.Path()):
                assert(True)
            case x:
                assert(False), x

    def test_coerce_param_int(self, exp):
        obj = DKey("{test!i}")
        state = {"test": "25"}
        assert(obj.data.convert == "i")
        match exp.expand(obj, state):
            case ExpInst_d(value=25):
                assert(True)
            case x:
                assert(False), x

    def test_coerce_param_fail(self, exp):
        obj = DKey("{test!i}")
        state = {"test": "blah"}
        assert(obj.data.convert == "i")
        assert(obj.data.convert in EXPANSION_CONVERT_MAPPING)
        with pytest.raises(ValueError):
            EXPANSION_CONVERT_MAPPING["i"]("blah")

        with pytest.raises(ValueError):
            exp.expand(obj, state)

    def test_multi_coerce_to_path(self, exp):
        obj = DKey("{test}/{test}", ctor=pl.Path)
        assert(DKey.MarkOf(obj) is list)
        state = {"test": "blah"}
        match exp.expand(obj, state):
            case ExpInst_d(value=pl.Path()):
                assert(True)
            case x:
                assert(False), x

    def test_multi_coerce_subkey(self, exp):
        obj = DKey("{test!p}/{test}")
        assert(DKey.MarkOf(obj) is list)
        assert(obj.keys()[0].data.convert == "p")
        state = {"test": "blah"}
        match exp.expand(obj, state):
            case ExpInst_d(value=str() as x):
                assert(x == str(pl.Path.cwd() / "blah/blah"))
                assert(True)
            case x:
                assert(False), x

    def test_multi_coerce_multi_subkey(self, exp):
        obj    = DKey("{test!p} : {test!p}")
        target = "".join(map(str, [(pl.Path.cwd() / "blah"), " : ", (pl.Path.cwd() / "blah")]))
        assert(DKey.MarkOf(obj) is list)
        assert(obj.keys()[0].data.convert == "p")
        state = {"test": "blah"}
        match exp.expand(obj, state):
            case ExpInst_d(value=str() as x):
                assert(x == target)
                assert(True)
            case x:
                assert(False), x

@pytest.mark.parametrize("ctor", Expanders)
class TestFallbacks:

    @pytest.fixture(scope="function")
    def exp(self, ctor):
        return ctor(ctor=DKey)

    def test_sanity(self, exp):
        assert(True is not False) # noqa: PLR0133

    def test_basic_key_fallback(self, exp):
        key = DKey("blah", implicit=True, fallback="aweg")
        match exp.expand(key):
            case ExpInst_d(value="aweg"):
                assert(True)
            case x:
                 assert(False), x

    def test_basic_expansion_fallback(self, exp):
        key = DKey("blah", implicit=True)
        match exp.expand(key, fallback="aweg"):
            case ExpInst_d(value="aweg"):
                assert(True)
            case x:
                 assert(False), x

    def test_fallback_typecheck(self, exp):
        key = DKey("blah", implicit=True, fallback="aweg", check=str)
        match exp.expand(key):
            case ExpInst_d(value="aweg"):
                assert(True)
            case x:
                 assert(False), x

    @pytest.mark.parametrize("fallback", [list, dict, set])
    def test_fallback_type_factory(self, exp, fallback):
        key = DKey("blah", implicit=True, fallback=fallback)
        match exp.expand(key):
            case ExpInst_d(value=list()|dict()|set()):
                assert(True)
            case x:
                 assert(False), x



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import re
import logging as logmod

# ##-- end stdlib imports

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from logging import LogRecord
    from jgdv import Maybe, RxStr
## isort: on
# ##-- end type checking

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# True to process, False to reject

class SimpleFilter:
    """
      A Simple filter to reject based on:
      1) a whitelist of regexs,
      2) a simple list of rejection names

    """

    def __init__(self, allow:Maybe[list[RxStr]]=None, reject:Maybe[list[str]]=None) -> None:
        self.allowed    = allow or []
        self.rejections = reject or []
        self.allowed_re    = re.compile("^({})".format("|".join(self.allowed)))
        if bool(self.allowed):
            msg = "Logging Allows are not implemented yet"
            raise NotImplementedError(msg)

    def __call__(self, record:LogRecord) -> bool:
        if record.name in ["root", "__main__"]:
            return True
        if not (bool(self.allowed) or bool(self.rejections)):
            return True

        rejected = False
        rejected |= any(x in record.name for x in self.rejections)
        return not rejected



#!/usr/bin/env python3
"""



"""
# Import:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import weakref
from uuid import UUID, uuid1
# ##-- end stdlib imports

from jgdv import Maybe
from jgdv.decorators import Decorator

from ._interface import Logger, LOGDEC_PRE

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard, Concatenate
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, Lambda
## isort: on
# ##-- end type checking

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Global Vars:

# Body:

class LogCall(Decorator):
    """ A Decorator for announcing the entry/exit of a function call

    eg:
    @LogCall(enter="Entering", exit="Exiting", level=logmod.INFO)
    def a_func()...
    """

    def __init__(self, enter:Maybe[str|Lambda]=None, exit:Maybe[str|Lambda]=None, level:int|str=logmod.INFO, logger:Maybe[Logger]=None) -> None:  # noqa: A002
        super().__init__(prefix=LOGDEC_PRE)
        self._logger = logger or logging
        self._enter_msg = enter
        self._exit_msg = exit
        match level:
            case str():
                self._level = logmod.getLevelNamesMapping().get(level, logmod.INFO)
            case int():
                self._level = level
            case _:
                raise ValueError(level)

    def _log_msg(self, msg:Maybe[str|Lambda], fn:Callable, args:Iterable, **kwargs:Any) -> None:  # noqa: ANN401
        match msg:
            case None:
                return None
            case types.FunctionType():
                msg = msg(fn, *args, **kwargs)
            case str():
                pass
            case _:
                raise TypeError(msg)

        self._logger.log(self._level, msg)

    def _wrap_method[X, **I, O](self, fn:Callable[Concatenate[X, I],O]) -> Callable[Concatenate[X, I],O]:

        def basic_wrapper(_self:X, *args:I.args, **kwargs:I.kwargs) -> O:
            self._log_msg(self._enter_msg, fn, args, obj=_self, **kwargs)
            ret_val = fn(_self, *args, **kwargs)
            self._log_msg(self._exit_msg, fn, args, obj=_self, returned=ret_val, **kwargs)
            return ret_val

        return basic_wrapper

    def _wrap_fn[**I, O](self, fn:Callable[I, O]) -> Callable[I, O]:

        def basic_wrapper(*args:I.args, **kwargs:I.kwargs) -> O:
            self._log_msg(self._enter_msg, fn, args, obj=None, **kwargs)
            ret_val = fn(*args, **kwargs)
            self._log_msg(self._exit_msg, fn, args, obj=None, returned=ret_val, **kwargs)
            return ret_val

        return basic_wrapper

    def _wrap_class[T](self, cls:type[T]) -> type[T]:
        raise NotImplementedError()



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from ._interface import Logger
    from jgdv import Maybe, Traceback
## isort: on
# ##-- end type checking


##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class LogContext:
    """
      a really simple wrapper to set a logger's level, then roll it back

    use as:
    with LogContext(logger, level=logmod.INFO) as ctx:
    ctx.log("blah")
    # or
    logger.info("blah")
    """

    def __init__(self, logger:Logger, level:Maybe[int]=None) -> None:
        self._logger          = logger
        self._original_level  = self._logger.level
        self._level_stack     = [self._original_level]
        self._temp_level      = level or self._original_level

    def __call__(self, level:int) -> Self:
        self._temp_level = level
        return self

    def __enter__(self) -> Self:
        match self._temp_level:
            case int() | str():
                self._level_stack.append(self._logger.level)
                self._logger.setLevel(self._temp_level)
        return self

    def __exit__(self, etype:Maybe[type], err:Maybe[Exception], tb:Maybe[Traceback]) -> bool:
        if bool(self._level_stack):
            self._logger.setLevel(self._level_stack.pop())
        else:
            self._logger.setLevel(self._original_level)
        if etype is None:
            return False

        return True

    def __getattr__(self, key:str) -> Logger:
        return cast("Logger", getattr(self._logger, key))

    def log(self, msg:str, *args:Any, **kwargs:Any) -> None:  # noqa: ANN401
        self._logger.log(self._temp_level, msg, *args, **kwargs)

class TempLogger:
    """ For using a specific type of logger in a context, or getting
    a custom logger class without changing it globally

    use as:
    with TempLogger(MyLoggerClass) as ctx:
    # Either:
    ctx['name'].info(...)
    # or:
    logmod.getLogger('name').info(...)
    """
    _target_cls  : type[Logger]
    _original    : Maybe[type[Logger]]

    def __init__(self, logger:type[Logger]) -> None:
        self._target_cls  = logger
        self._original    = None

    def __enter__(self) -> Self:
        self._original = logmod.getLoggerClass()
        logmod.setLoggerClass(self._target_cls)
        return self

    def __exit__(self, etype:Maybe[type], err:Maybe[Exception], tb:Maybe[Traceback]) -> bool:
        if self._original is not None:
            logmod.setLoggerClass(self._original)
        if etype is None:
            return False

        return True



#!/usr/bin/env python3
"""

"""
# Imports:
from __future__ import annotations

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, Frame
## isort: on
# ##-- end type checking

# Vars:
DEL_LOG_K         : Final[str] = "_log_del_active"
DEFAULT_LOG_LEVEL : Final[int] = 40
DEFAULT_REPORT    : Final[str] = "traceback"
CHANGE_PREFIX     : Final[str] = "+-:"
DEFAULT_PREFIX    : Final[str] = "--"

type TraceEvent = (
    Literal["call"]
    | Literal["line"]
    | Literal["return"]
    | Literal["exception"]
    | Literal["opcode"]
)
# Body:

class TraceFn_p(Protocol):
    """ The protocol for functions for sys.settrace """

    def __call__(self, frame:Frame, event:TraceEvent, arg:Any) -> Maybe[TraceFn_p]: ... # noqa: ANN401



#!/usr/bin/env python3
"""

"""
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import collections
import contextlib
import hashlib
from copy import deepcopy
from uuid import UUID, uuid1
import atexit # for @atexit.register
import faulthandler
# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

from dataclasses import dataclass, field, InitVar
from typing import Any

if TYPE_CHECKING:
    import types
    from jgdv import Maybe, Rx
    from typing import Final
    from typing import ClassVar, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
DEFAULT_PREFIX  : Final[str]  = "-"
END_SEP         : Final[str]  = "--"
FULLNAME_RE     : Final[Rx]   = re.compile(r"(?:<(?P<pos>\d*)>|(?P<prefix>\W+))?(?P<name>.+?)(?P<assign>=)?$")
DEFAULT_DOC     : Final[str]  = "A Base Parameter"
""" The Regexp for parsing string descriptions of parameters """

##--|
EMPTY_CMD           : Final[str]  = "_cmd_"
EXTRA_KEY           : Final[str]  = "_extra_"
NON_DEFAULT_KEY     : Final[str]  = "_non_default_"
DEFAULT_COUNT       : Final[int]  = 1
UNRESTRICTED_COUNT  : Final[int]  = -1
##--|
TYPE_CONV_MAPPING: Final[dict[str|type|types.GenericAlias, type|Callable]] = {
    "int"               : int,
    "float"             : float,
    "bool"              : bool,
    "str"               : str,
    "list"              : list,
}

# Body:

class SectionType_e(enum.Enum):
    """ The different types of section a parse machine can process """
    prog = enum.auto()
    cmd  = enum.auto()
    sub  = enum.auto()

class ParseResult_d:
    """ Simple container for parsed cli information

    Args:
        name : the name of the spec that parsed this data
        args : mapping of {arg -> data}, including default values it nothing parsing for it
        non_default : set[arg, ...] that actually parsed
        ref : the name of a linked parse result this object extends

    """
    __slots__ = ("args", "name", "non_default", "ref")
    name         : str
    args         : dict
    non_default  : set[str]
    ref          : Maybe[str]

    def __init__(self, name:str, args:Maybe[dict]=None, ref:Maybe[str]=None) -> None:
        self.name         = name
        self.args         = args or {}
        self.non_default  = set()
        self.ref          = ref

    @override
    def __repr__(self) -> str:
        return f"<ParseResult: {self.name}, args:{self.args}>"

    def to_dict(self) -> dict:
        return {"name":self.name, "args":self.args, NON_DEFAULT_KEY:self.non_default}

class ParseReport_d:
    """ The returned data of parsing cli args

    Args:
        raw : the raw args that were used. ieg: sys.argv[:]
        remaining : anything not parsed
        prog : ParseResult_d of base program arguments
        cmds : mapping(cmdName -> [ParseResult_d])
        subs : mapping(subName -> [ParseResult_d])
        help : bool

    """
    type CmdName = str
    type SubName = str
    __slots__ = ("cmds", "help", "prog", "raw", "remaining", "subs")
    raw        : tuple[str, ...]
    remaining  : tuple[str, ...]
    prog       : ParseResult_d
    cmds       : dict[CmdName, tuple[ParseResult_d]]
    subs       : dict[SubName, tuple[ParseResult_d]]
    help       : bool

    def __init__(self, *, raw:Iterable[str], remaining:Iterable[str], prog:ParseResult_d, _help:bool) -> None:
        self.raw        = tuple(raw)
        self.remaining  = tuple(remaining)
        self.help       = _help
        self.prog       = prog
        self.cmds       = {}
        self.subs       = {}

    def to_dict(self) -> dict:
        return {
            "prog" : self.prog.to_dict(),
            "cmds" : {x: [y.to_dict() for y in ys] for x,ys in self.cmds.items()},
            "subs" : {x: [y.to_dict() for y in ys] for x,ys in self.subs.items()},
            "help" : self.help,
        }

##--| Params

@runtime_checkable
class ParamSpec_p(Protocol):
    """ Base class for CLI param specs, for type matching
    when 'consume' is given a list of strs,
    it can match on the args,
    and return an updated diction and a list of values it didn't consume

    """

    @classmethod
    def key_func(cls, x:ParamSpec_i) -> tuple: ...

    def consume(self, args:list[str], *, offset:int=0) -> Maybe[tuple[dict, int]]:
        pass

    ##--| properties

    @property
    def short(self) -> str: ...

    @property
    def inverse(self) -> str: ...

    @property
    def repeatable(self) -> bool: ...

    @property
    def key_str(self) -> str: ...

    @property
    def short_key_str(self) -> Maybe[str]: ...

    @property
    def key_strs(self) -> list[str]: ...

    @property
    def default_value(self) -> Any: ...  # noqa: ANN401

    @property
    def default_tuple(self) -> tuple[str, Any]: ...

    def help_str(self, *, force:bool=False) -> str: ...

class ParamSpec_i(ParamSpec_p, Protocol):
    _processor : ClassVar

    name       : str
    type_      : Maybe[type]
    insist     : bool
    default    : Any|Callable
    desc       : str
    count      : int
    prefix     : int|str
    separator  : str|Literal[False]
    implicit   : bool

##--| Param Subtypes

@runtime_checkable
class PositionalParam_p(Protocol):
    """ Interface to mark a parameter as positional """

    def _positional(self) -> Literal[True]: ...

@runtime_checkable
class AssignmentParam_p(Protocol):
    """ Interface to mark a parameter as an assignment """

    def _assignment(self) -> Literal[True]: ...

@runtime_checkable
class KeyParam_p(Protocol):
    """ Interface to mark a parameter as a key value pair """

    def _keyval(self) -> Literal[True]: ...

@runtime_checkable
class ToggleParam_p(Protocol):
    """ Interface to mark a parameter as a boolean toggle """

    def _toggle(self) -> Literal[True]: ...
##--| Parsing

@runtime_checkable
class ArgParserModel_p(Protocol):
    """ The Model used in a jgdv.cli.arg_parser:ParseMachine to implement specific parsing logic """

    def prepare_for_parse(self, *, prog:ParamSource_p, cmds:list[ParamSource_p], subs:list[tuple[tuple[str, ...], ParamSource_p]], raw_args:list[str]) -> None: ...

@runtime_checkable
class ParamSource_p(Protocol):
    """ Param Sources are anything that can provide a name and a set of parameters """

    @property
    def name(self) -> str:
        raise NotImplementedError()

    def param_specs(self) -> list[ParamSpec_i]:
        raise NotImplementedError()

@runtime_checkable
class CLIParamProvider_p(Protocol):
    """
      Things that can provide parameter specs for CLI parsing
    """

    @classmethod
    def param_specs(cls) -> list[ParamSpec_i]:
        """  make class parameter specs  """
        pass



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import re
import time
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv.files.bookmarks.bookmark import Bookmark

# ##-- end 1st party imports

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
from pydantic import BaseModel, Field, model_validator, field_validator, ValidationError
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    import pathlib as pl
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe
## isort: on
# ##-- end type checking


##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class BookmarkCollection(BaseModel):
    """A container of bookmarks,
    read from a file where each line is a bookmark url with tags.
    """

    entries : list[Bookmark] = []
    ext     : str            = ".bookmarks"

    @staticmethod
    def read(fpath:pl.Path) -> BookmarkCollection:
        """ Read a file to build a bookmark collection """
        bookmarks = BookmarkCollection()
        for line in (x.strip() for x in fpath.read_text().split("\n")):
            if not bool(line):
                continue
            bookmarks += Bookmark.build(line)

        return bookmarks

    @override
    def __str__(self) -> str:
        return "\n".join(map(str, sorted(self.entries)))

    @override
    def __repr__(self) -> str :
        return f"<{self.__class__.__name__}: {len(self)}>"

    def __iadd__(self, value:Bookmark) -> Self:
        return self.update(value)

    @override
    def __iter__(self) -> Iterator[Bookmark]: # type: ignore[override]
        return iter(self.entries)

    def __contains__(self, value:Bookmark) -> bool:
        return value in self.entries

    def __len__(self) -> int:
        return len(self.entries)

    @override
    def __hash__(self) -> int:
        return id(self)

    def update(self, *values:Bookmark|BookmarkCollection|Iterable) -> Self:
        for val in values:
            match val:
                case Bookmark():
                    self.entries.append(val)
                case BookmarkCollection():
                    self.entries += val.entries
                case [*vals] | set(vals):
                    self.update(*vals)
                case _:
                    raise TypeError(type(val))
        return self

    def difference(self, other:Self) -> BookmarkCollection:
        result = BookmarkCollection()
        for bkmk in other:
            if bkmk not in self:
                result += bkmk

        return result

    def merge_duplicates(self) -> None:
        deduplicated = {}
        for x in self:
            if x.url not in deduplicated:
                deduplicated[x.url] = x
            else:
                deduplicated[x.url] = x.merge(deduplicated[x.url])

        self.entries = list(deduplicated.values())



#!/usr/bin/env python3
"""

"""
# Import:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import weakref
from uuid import UUID, uuid1
# ##-- end stdlib imports

import stackprinter

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe, RxStr
## isort: on
# ##-- end type checking

# Global Vars:

# Body:

class StackFormatter_m:
    """ A Mixin Error formatter, adapted from stackprinter's docs
    Compactly Formats the error stack trace, without src.

    """

    indent_str       : ClassVar[str]         = "  |  "
    suppress         : ClassVar[list[RxStr]] = [r".*pydantic.*", r"<frozen importlib._bootstrap>"]
    source_height    : ClassVar[int]         = 10
    source_lines     : ClassVar[int|str]     = 0
    use_stackprinter : bool                  = True

    def formatException(self, exc_info:Maybe[tuple]) -> str:  # noqa: N802
        match exc_info:
            case None | (None, None, None):
                return ""
            case _ if not self.use_stackprinter:
                return cast("str", super().formatException(exc_info)) # type: ignore[misc]

        msg : str = stackprinter.format(exc_info,
                                        source_lines=self.source_lines,
                                        suppressed_paths=self.suppress)
        lines = [x for x in msg.splitlines() if bool(x)]
        indented = [f"{self.indent_str}{line}\n" for line in lines[-self.source_height:]]
        return "".join(indented)

    def formatStack(self, stack_info:str) -> str:  # noqa: N802
        return cast("str", super().formatStack(stack_info)) # type: ignore[misc]



#!/usr/bin/env python3
"""

"""
# Import:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import weakref
from uuid import UUID, uuid1
# ##-- end stdlib imports

from ._interface import LogLevel_e

# ##-- types
# isort: off
# General
import abc
import collections.abc
import typing
import types
from typing import cast, assert_type, assert_never
from typing import Generic, NewType, Never
from typing import no_type_check, final, override, overload
# Protocols and Interfaces:
from typing import Protocol, runtime_checkable
# isort: on
# ##-- end types

# ##-- type checking
# isort: off
if typing.TYPE_CHECKING:
    from typing import Final, ClassVar, Any, Self
    from typing import Literal, LiteralString
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv import Maybe
## isort: on
# ##-- end type checking

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
LoggerClass : Final[type] = logmod.getLoggerClass()
# Body:

class JGDVLogger(LoggerClass):
    """ Basic extension of the logger class

    checks the classvar _levels (intEnum) for additional log levels
    which can be accessed as attributes and items.
    eg: logger.trace(...)
    and: logger['trace'](...)


    A Logger can add prefixes to a logged messages.
    eg:_

    ..code: python

        logger.set_prefixes('[Test]')
        logger.info('this is a test message')
        # Result :  '[Test] this is a test message'

    """
    _prefixes  : list[str|Callable]
    _colour    : Maybe[str]

    @classmethod
    def install(cls) -> None:
        logmod.setLoggerClass(cls)

    def __init__(self, *args:Any, **kwargs:Any) -> None:  # noqa: ANN401
        super().__init__(*args, **kwargs)
        self._prefixes = []
        self._colour   = None

    def __getattr__(self, attr:str) -> Callable:
        try:
            return ftz.partial(self.log, LogLevel_e[attr])
        except KeyError:
            msg = "Invalid Extension Log Level"
            raise AttributeError(msg, attr) from None

    def __getitem__(self, key:str) -> Callable:
        return self.__getattr__(key)

    ##--| public methods
    def set_colour(self, colour:Maybe[str]) -> None:
        self._colour = self._colour or colour

    def set_prefixes(self, *prefixes:str|Callable) -> None:
        """
        Set prefixes for the logger to add to logged messages
        """
        match prefixes:
            case None | ():
                pass
            case [*xs]:
                self._prefixes =  list(xs)

    def prefix(self, prefix:str|Callable) -> Self:
        """ Create a new logger, with a prefix """
        match prefix:
            case str():
                child = self.getChild(prefix)
            case x if callable(x):
                child = self.getChild(prefix.__name__)
            case _:
                raise TypeError(prefix)

        child.set_prefixes(*self._prefixes, prefix)
        return child

    def getChild(self, name:str) -> JGDVLogger:  # noqa: N802
        """
        Create a child logger, copying the colour of this logger
        """
        child = cast("JGDVLogger", super().getChild(name))
        child.set_colour(self._colour)
        return child

    def makeRecord(self, *args:Any, **kwargs:Any) -> logmod.LogRecord:  # noqa: ANN401, N802
        """
        A factory method which can be overridden in subclasses to create
        specialized LogRecords.
        args: name, level, fn, lno, msg, args, exc_info,
        kwargs: func=None, extra=None, sinfo=None
        """
        rv        : logmod.LogRecord
        modified  : list  = list(args)
        msg_total         = []
        for pre in self._prefixes:
            match pre:
                case None:
                    pass
                case str():
                    msg_total.append(pre)
                case x if callable(x):
                    msg_total.append(x())
        else:
            match args[4]:
                case str():
                    msg_total.append(args[4])
                case x:
                    msg_total.append("%s")
                    modified[5] = [args[4],  *args[5]]
            modified[4] = "".join(msg_total)

        rv = super().makeRecord(*modified, **kwargs)
        if self._colour and "colour" not in rv.__dict__:
            rv.__dict__["colour"] = self._colour
        return rv
